//
// MigraDoc - Creating Documents on the Fly
//
// This is a part of the empira MigraDoc Document Object Model.
//
// Copyright (c) 2002-2005 empira Software GmbH, Cologne (Germany)
// All rights reserved
//
// http://www.pdfsharp.com
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

using System;
using System.Collections;
using MigraDoc.DocumentObjectModel.Internals;
using MigraDoc.DocumentObjectModel.IO;

namespace MigraDoc.DocumentObjectModel.Visitors
{
  /// <summary>
  /// Flattens a document for PDF rendering.
  /// </summary>
  public class PdfFlattenVisitor : RtfFlattenVisitor
  {
    /// <summary>
    /// Initializes a new instance of the PdfFlattenVisitor class.
    /// </summary>
    /// <param name="docObj">The document object to flatten. Usually the Document itself.</param>
    public PdfFlattenVisitor(DocumentObject docObj) : base(docObj)
    {
    }
    
    internal override void VisitDocumentElements (DocumentElements elements)
    {
      SortedList splitParaList = new SortedList();

      for(int idx = 0; idx < elements.Count; ++idx)
      {
        Paragraph paragraph = elements[idx] as Paragraph;
        if (paragraph != null)
        {
          Paragraph[] paragraphs = paragraph.SplitOnParaBreak();
          if (paragraphs != null)
            splitParaList.Add(idx, paragraphs);
        }
      }

      int insertedObjects = 0;
      for(int idx = 0; idx < splitParaList.Count; ++idx)
      {
        int insertPosition = (int)splitParaList.GetKey(idx);
        Paragraph[] paragraphs = (Paragraph[])splitParaList.GetByIndex(idx);
        foreach(Paragraph paragraph in paragraphs)
        {
          elements.InsertObject(insertPosition + insertedObjects, paragraph);
          ++insertedObjects;
        }
        elements.RemoveObjectAt(insertPosition + insertedObjects);
        --insertedObjects;
      }
    }

    internal override void VisitDocumentObjectCollection(DocumentObjectCollection elements)
    {
      ArrayList textIndices = new ArrayList();
      if (elements is ParagraphElements)
      {
        for (int idx = 0; idx < elements.Count; ++idx)
        {
          if (elements[idx] is Text)
            textIndices.Add(idx);
        }
      }

      int[] indices = (int[])textIndices.ToArray(typeof(int));
      int insertedObjects = 0;
      foreach (int idx in indices)
      {
        Text text = (Text)elements[idx + insertedObjects];
        string currentString = "";
        foreach (char ch in text.Content)
        {
          switch (ch)
          {
            case ' ':
            case '\r':
            case '\n':
            case '\t':
              if (currentString != "")
              {
                elements.InsertObject(idx + insertedObjects, new Text(currentString));
                ++insertedObjects;
                currentString = "";
              }
              elements.InsertObject(idx + insertedObjects, new Text(" "));
              ++insertedObjects;
              break;
            
            case '-': //minus
              elements.InsertObject(idx + insertedObjects, new Text(currentString + ch));
              ++insertedObjects;
              currentString = "";
              break;

            case '': //soft hyphen
              if (currentString != "")
              {
                elements.InsertObject(idx + insertedObjects, new Text(currentString));
                ++insertedObjects;
                currentString = "";
              }
              elements.InsertObject(idx + insertedObjects, new Text(""));
              ++insertedObjects;
              currentString = "";
              break;

            default:
              currentString += ch;
              break;
          }
        }
        if (currentString != "")
        {
          elements.InsertObject(idx + insertedObjects, new Text(currentString));
          ++insertedObjects;
        }
        elements.RemoveObjectAt(idx + insertedObjects);
        --insertedObjects;
      }
    }



    internal override void VisitFormattedText(FormattedText formattedText)
    {
      base.VisitFormattedText(formattedText);

      Font parentFont = GetParentFont(formattedText);

      if (formattedText.font == null)
        formattedText.Font = parentFont.Clone();
      else if (parentFont != null)
        FlattenFont(formattedText.font, parentFont);
    }
  }
}
