//
// MigraDoc - Creating Documents on the Fly
//
// Authors:
//   Klaus Potzesny (mailto:Klaus.Potzesny@pdfsharp.com)
//
// This is a part of the empira MigraDoc RtfRendering module.
//
// Copyright (c) 2002-2005 empira Software GmbH, Cologne (Germany)
// All rights reserved
//
// http://www.pdfsharp.com
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

using System;
using System.IO;
using System.Text;

namespace MigraDoc.RtfRendering
{
  /// <summary>
  /// Class to write RTF output.
  /// </summary>
  public class RtfWriter
  {
    /// <summary>
    /// Initializes a new instance of the RtfWriter class.
    /// </summary>
    public RtfWriter(TextWriter textWriter)
    {
      this.textWriter = textWriter;
    }

    /// <summary>
    /// Writes a left brace.
    /// </summary>
    public void StartContent()
    {
      this.textWriter.Write("{");
      this.lastWasControl = false;
    }

    /// <summary>
    /// Writes a right brace.
    /// </summary>
    public void EndContent()
    {
      this.textWriter.Write("}");
      this.lastWasControl = false;
    }
    
    
    /// <summary>
    /// Writes the given text, handling special characters before.
    /// </summary>
    public void WriteText(string text)
    {
      StringBuilder strBuilder = new StringBuilder(text.Length);
      if (this.lastWasControl)
        strBuilder.Append(" ");

      foreach (char ch in text)
      {
        switch(ch)
        {
          case '\\':
            strBuilder.Append(@"\\");
            break;

          case '{':
            strBuilder.Append(@"\{");
            break;

          case '}':
            strBuilder.Append(@"\}");
            break;

          case '': //character 173, softhyphen
            strBuilder.Append(@"\-");
            break;
            
          //REM: Are there any further special characters?
          default:
            strBuilder.Append(ch);
            break;
        }
      }
      this.textWriter.Write(strBuilder.ToString());
      lastWasControl = false;
    }

    /// <summary>
    /// Writes the number as hex value. only numbers &lt;= 255 are allowed.
    /// </summary>
    public void WriteHex(uint hex)
    {
      if (hex > 0xff)
//TODO: Fehlermeldung
        return;

      this.textWriter.Write(@"\'" + hex.ToString("x"));
      lastWasControl = false;
      //Dahinter darf kein zustzliches Blank stehen.
    }

    /// <summary>
    /// Writes a blank in paragraph text.
    /// </summary>
    public void WriteBlank()
    {
      this.textWriter.Write(" ");
    }

    /// <summary>
    /// Writes a semicolon as separator e.g in in font tables.
    /// </summary>
    public void WriteSeparator()
    {
      this.textWriter.Write(";");
      lastWasControl = false;
    }


    /// <summary>
    /// Writes the given string as control word optionally with a star.
    /// </summary>
    public void WriteControl(string ctrl, bool withStar)
    {
      if (!withStar)
       WriteControl(ctrl);
      else
        WriteControlWithStar(ctrl);
    }

    /// <summary>
    /// Writes the given string as control word with a star followed by a space.
    /// </summary>
    public void WriteControl(string ctrl, string value, bool _withStar)
    {
      if (_withStar)
        WriteControlWithStar(ctrl, value);
      else
        WriteControl(ctrl, value);
    }

    /// <summary>
    /// Writes the given string as control word with a star followed by a space.
    /// </summary>
    public void WriteControl(string ctrl, int value, bool withStar)
    {
      WriteControl(ctrl, value.ToString(), withStar);
    }

    /// <summary>
    /// Writes the given string as control word.
    /// </summary>
    public void WriteControl(string ctrl)
    {
      WriteControl(ctrl, "");
    }

    /// <summary>
    /// Writes the given string and integer as control word / value pair followed by a space.
    /// </summary>
    public void WriteControl(string ctrl, int value)
    {
      WriteControl(ctrl, value.ToString());
    }

    /// <summary>
    /// Writes the given strings as control word / value pair.
    /// </summary>
    public void WriteControl(string ctrl, string value)
    {
      this.textWriter.Write("\\" + ctrl + value);
      lastWasControl = true;
    }

    /// <summary>
    /// Writes the given string and integer as control word / value pair
    /// with a star.
    /// </summary>
    public void WriteControlWithStar(string ctrl)
    {
      WriteControlWithStar(ctrl, "");
    }

    /// <summary>
    /// Writes the given string and integer as control word / value pair followed by a space.
    /// </summary>
    public void WriteControlWithStar(string ctrl, int value)
    {
      WriteControlWithStar(ctrl, value.ToString());
    }

    /// <summary>
    /// Writes the given string and integer as control word / value pair
    /// with a star.
    /// </summary>
    public void WriteControlWithStar(string ctrl, string value)
    {
      this.textWriter.Write("\\*\\" + ctrl + value);
      lastWasControl = true;
    }

    private bool lastWasControl = false;
    private TextWriter textWriter;
  }
}
